#define _XOPEN_SOURCE 1
#define _XOPEN_SOURCE_EXTENDED 1
#include <stddef.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <termios.h>
#include <unistd.h>
#include <errno.h>

static char const prog[]="cvmlogin";
#define service (prog+3)

static int read_line(int const fd, char* buf, size_t size,
                     size_t* const plen) {
  size_t len=0;
  while (size!=0) {
    ssize_t const nread=read(fd, buf, 1);
    if (nread<0) {
      if (errno==EINTR) continue;
      return -1;
    }
    if (nread==0) break;
    --size;
    ++len;
    if (*(buf++)=='\n') break;
  }
  *plen=len;
  return 0;
}

static buffer* logbuf;
static int status=111;

void log(char const* const x0, char const* const x1) {
  do {
    if (buffer_puts(logbuf, service)!=0) break;
    if (buffer_puts(logbuf, err_)!=0) break;
    if (buffer_puts(logbuf, x0)!=0) break;
    if (x1!=null)
      if (buffer_puts(logbuf, x1)!=0) break;
    if (buffer_putsflush(logbuf, "\n")!=0) break;
    return;
  } while (0);
  _exit(status);
}

void die(char const* const x0, char const* const x1, char const* const x2) {
  char const* const err=(errno==0? null: error_str(errno));
  buffer* bufs[]={ null, null, null };
  buffer** pbuf=bufs;
  bufs[0]=buffer_1;
  bufs[1]=buffer_2;
  for (;; ++pbuf) {
    buffer* const buf=*pbuf;
    if (buf==null) break;
    if (buffer_puts(buf, prog)!=0) continue;
    if (buffer_puts(buf, fatal)!=0) continue;
    if (buffer_puts(buf, x0)!=0) continue;
    if (x1!=null) {
      if (buffer_puts(buf, x1)!=0) continue;
      if (x2!=null)
        if (buffer_puts(buf, x2)!=0) continue;
    }
    if (err!=null) {
      if (buffer_puts(buf, err_)!=0) continue;
      if (buffer_puts(buf, err)!=0) continue;
    }
    buffer_putsflush(buf, "\n");
  }
  _exit(status);
}

static void check(char* const buf, size_t const len, size_t const max,
                  char const* const name) {
  char const* err=null;
  if (len==0 || buf[len-1]!='\n') {
    errno=0;
    die("EOF while reading ", name, null);
  }
  if (len==max) err=" too long";
  else if (byte_chr(buf, len, '\0')!=len) err=" contains null characters";
  buf[len-1]='\0';
  if (err==null) return;
  errno=0;
  die(name, err, null);
}

static char const var_cvm[]="$CVM_PLAIN";

int main(int argc, char** argv) {
  char const* x;
  typedef char buf[512];
  static size_t const bufmax=(sizeof (buf))-1;
  buf ubuf, pbuf;
  size_t ulen, plen;
  struct termios normal, noecho;
#ifdef TCSASOFT
  int const tcsa=TCSAFLUSH|TCSASOFT;
#else
  int const tcsa=TCSAFLUSH;
#endif
  int ret, err;
  char const* const cvm=env_get(var_cvm+1);
  struct stralloc sa={ 0 };

  logbuf=buffer_2;

  if (cvm==null) {
    errno=0;
    die(var_cvm, err_notset, null);
  }
  if (tcgetattr(fileno_stdin, &normal)!=0)
    die(err_unable, err_tcgetattr, null);
  noecho=normal;
  noecho.c_lflag&=~(ECHO|ISIG);

  log("tty ", env_get("TTY"));

  x=env_get("PROMPT_LOGIN");
  if (x==null) x="login: ";
  if (buffer_putsflush(buffer_1, x)!=0)
    die(err_unable, "write login prompt", null);
  if (read_line(fileno_stdin, ubuf, bufmax, &ulen)!=0)
    die(err_unable, "read username", null);
  check(ubuf, ulen, bufmax, "username");

  log("attempt ", ubuf);

  if (env_get("DENY")!=null) {
    if (stralloc_copys(&sa, "ALLOW_")==0 ||
        stralloc_cats(&sa, ubuf)==0 ||
        stralloc_append(&sa, "")==0)
      die(err_unable, err_alloc, null);
    err=(env_get(sa.s)==null);
  } else {
    if (stralloc_copys(&sa, "DENY_")==0 ||
        stralloc_cats(&sa, ubuf)==0 ||
        stralloc_append(&sa, "")==0)
      die(err_unable, err_alloc, null);
    err=(env_get(sa.s)!=null);
  }
  if (err) {
    x="login denied";
    status=100;
    log(x, null);
    logbuf=buffer_1;
    log(x, null);
    _exit(status);
  }

  x=env_get("PROMPT_PASSWORD");
  if (x==null) x="Password: ";
  if (buffer_putsflush(buffer_1, x)!=0)
    die(err_unable, "write password prompt", null);
  if (tcsetattr(fileno_stdin, tcsa, &noecho)!=0)
    die(err_unable, err_tcsetattr, null);
  ret=read_line(fileno_stdin, pbuf, bufmax, &plen);
  err=errno;
  if (tcsetattr(fileno_stdin, tcsa, &normal)!=0)
    die(err_unable, err_tcsetattr, null);
  if (ret!=0) {
    errno=err;
    die(err_unable, "read password", null);
  }
  if (buffer_putsflush(buffer_1, "\n")!=0)
    die(err_unable, "write newline", null);
  check(pbuf, plen, bufmax, "password");

  {
    char const* credentials[]={ null, null };
    credentials[0]=pbuf;
    ret=cvm_authenticate(cvm, ubuf, null, credentials, 0);
  }
  if (ret!=0) {
    if (ret==CVME_PERMFAIL) {
      x="login incorrect";
      status=100;
      log(x, null);
      logbuf=buffer_1;
      log(x, null);
      _exit(status);
    }
    switch (ret) {
      case CVME_GENERAL:     x="unspecified error";        break;
      case CVME_BAD_CLIDATA: x="invalid data from client"; break;
      case CVME_BAD_MODDATA: x="invalid data from module"; break;
      case CVME_IO:          x="I/O error";                break;
      default:               x=err_unknown;                break;
    }
    errno=0;
    die("error authenticating", err_, x);
  }

  log("login ok", null);

  {
    pid_t pid=fork();
    switch (pid) {
      case -1: die(err_unable, err_fork, null);
      case 0: break;
      default:
        pid=waitpid(pid, null, 0);
        err=errno;
        log("logout", null);
        if (pid>=0) _exit(0);
        errno=err;
        die(err_unable, err_wait, null);
    }
  }

  if (pathexec_env(var_user+1, cvm_fact_username)==0)
    die(err_unable, err_setvar, var_user);
  if (stralloc_copyb(&sa, "", 0)==0)
    die(err_unable, err_alloc, null);
  {
    char const* c;
    if (cvm_fact_str(CVM_FACT_USERID, &c)==0)
      if (pathexec_env(var_uid+1, c)==0)
        die(err_unable, err_setvar, var_uid);
    if (cvm_fact_str(CVM_FACT_GROUPID, &c)==0) {
      if (pathexec_env(var_gid+1, c)==0)
        die(err_unable, err_setvar, var_gid);
      if (env_get("INCLUDE_GID")!=null)
        if (stralloc_copys(&sa, " ")==0 ||
            stralloc_cats(&sa, c)==0)
          die(err_unable, err_alloc, null);
    }
    while (cvm_fact_str(CVM_FACT_SUPP_GROUPID, &c)==0)
      if (stralloc_cats(&sa, " ")==0 ||
          stralloc_cats(&sa, c)==0)
        die(err_unable, err_alloc, null);
  }
  x=env_get("EXTRA_GROUPS");
  if (x!=null && stralloc_cats(&sa, x)==0)
    die(err_unable, err_alloc, null);
  if (stralloc_append(&sa, "")==0)
    die(err_unable, err_alloc, null);
  if (pathexec_env(var_groups+1, sa.s)==0)
    die(err_unable, err_setvar, var_groups);
  if (pathexec_env(var_name+1, cvm_fact_realname)==0)
    die(err_unable, err_setvar, var_name);
  if (pathexec_env(var_home+1, cvm_fact_directory)==0)
    die(err_unable, err_setvar, var_home);
  if (pathexec_env(var_shell+1, cvm_fact_shell)==0)
    die(err_unable, err_setvar, var_shell);

  close(fileno_stderr);
  if (dup(fileno_stdout)!=fileno_stderr)
    _exit(111);

  {
    char* command[]={ (char*)PACKAGE_HOME "/command/setstate",
                      (char*)"lofgGuh",
                      (char*)PACKAGE_HOME "/command/loginshell",
                      null };
    if (argc>1) ++argv;
    else argv=command;
    pathexec((char const**)argv);
    die(err_unable, err_exec, argv[0]);
  }
}
